#ifndef FSM_BITS_AVAIABLE_STATES_HPP
#define FSM_BITS_AVAIABLE_STATES_HPP

#include <boost/mpl/fold.hpp>
#include <boost/mpl/front.hpp>
#include <boost/mpl/push_front.hpp>
#include <boost/mpl/pop_front.hpp>

#include <boost/mpl/if.hpp>
#include <boost/mpl/contains.hpp>
#include <boost/mpl/logical.hpp>
#include <boost/mpl/comparison.hpp>
#include <boost/mpl/empty.hpp>


// Graf bejarassal megkeressuk azokat az allapotokat, amelyek
// elerhetoek barmely vegpontbol, es azokat a szabalyokat kiszedjuk a
// szabaly halmazbol, amelyek ilyen allapotba vezetnek. Ezaltal
// a zsakutcakat eltavolitottuk az automatabol.

namespace fsm {

  namespace bits {

    using namespace boost::mpl;

    template <typename _Processed, typename _Reached, 
              typename _Transitions>
    struct AvailableSetStep {

      typedef typename front<_Reached>::type State;

      typedef typename push_front<_Processed, State>::type Processed;
      typedef typename pop_front<_Reached>::type PreReached;

      typedef typename 
      fold<_Transitions, 
           PreReached,
           if_<
             and_<
               is_same<State, Target<_2> >,
               and_<                 
                 not_<contains<Processed, Source<_2> > >,
                 not_<contains<PreReached, Source<_2> > > 
                 > >, 
             push_front<_1, Source<_2> >, _1> 
           >::type Reached;
           
    };

    template <typename _Processed, typename _Reached, 
              typename _Transitions, 
              bool _stop = empty<_Reached>::value>
      struct AvailableSetIteration {

      typedef AvailableSetStep<_Processed, _Reached, _Transitions> NextStep;
      
      typedef typename 
      AvailableSetIteration<typename NextStep::Processed, 
                            typename NextStep::Reached,
                            _Transitions>::type type;
      
    };
 

    template <typename _Processed, typename _Reached, 
              typename _Transitions>
    struct AvailableSetIteration<_Processed, _Reached, _Transitions, true> {
      typedef _Processed type;
    };
 

    template <typename _StopSet, typename _Transitions>
    struct AvailableStateSet {
      typedef typename 
      AvailableSetIteration<vector0<>, _StopSet, _Transitions>::type type;
    };

    template <typename _StopSet, typename _Transitions>
    struct AvailableTransitions {
      typedef typename AvailableStateSet<_StopSet, 
                                         _Transitions>::type AvailableStates;

      typedef typename 
      fold<_Transitions, vector0<>,
           if_<contains<AvailableStates, Target<_2> >, push_front<_1, _2>, _1> 
      >::type type;
    };
    
  }

}

#endif
