#ifndef FSM_BITS_EPSILON_EQUIVALENCE_HPP
#define FSM_BITS_EPSILON_EQUIVALENCE_HPP

#include <boost/mpl/fold.hpp>
#include <boost/mpl/front.hpp>
#include <boost/mpl/push_front.hpp>
#include <boost/mpl/pop_front.hpp>

#include <boost/mpl/if.hpp>
#include <boost/mpl/logical.hpp>
#include <boost/mpl/comparison.hpp>
#include <boost/mpl/empty.hpp>


// Graf bejarassal megkeressuk azokat az allapotokat, amelyek
// epszilon szabalyokon elerhetoek a kezdopontbol. Amikor egy
// atmenetet megvizsgalunk, akkor a vegallapot osszes ekvivalens
// celallapotat hozzaadjuk az uj allapothalmazhoz.

namespace fsm {

  namespace bits {

    using namespace boost::mpl;

    template <typename _Processed, typename _Reached, 
              typename _Transitions>
    struct EpsilonEquivalenceStep {

      typedef typename front<_Reached>::type State;

      typedef typename push_front<_Processed, State>::type Processed;
      typedef typename pop_front<_Reached>::type PreReached;
      
      typedef typename 
      fold<_Transitions, 
           PreReached,
           if_<
             and_<
               and_<
                 is_same<State, Source<_2> >,
                 is_same<Epsilon, Char<_2> >
               >,
               and_<                 
                 not_<contains<Processed, Target<_2> > >,
                 not_<contains<PreReached, Target<_2> > > 
               > 
             >, 
             push_front<_1, Target<_2> >, 
             _1
           > 
      >::type Reached;
      
    };

    template <typename _Processed, typename _Reached, 
              typename _Transitions, 
              bool _stop = empty<_Reached>::value>
      struct EpsilonEquivalenceIteration {
        
      typedef EpsilonEquivalenceStep<_Processed, 
                                     _Reached, _Transitions> NextStep;
        
      typedef typename 
      EpsilonEquivalenceIteration<typename NextStep::Processed, 
                                  typename NextStep::Reached,
                                  _Transitions>::type type;
      
    };
 

    template <typename _Processed, typename _Reached, 
              typename _Transitions>
    struct EpsilonEquivalenceIteration<_Processed, _Reached, 
                                       _Transitions, true> {
      typedef _Processed type;
    };
 

    template <typename _State, typename _Transitions>
    struct EpsilonEquivalence {
      typedef typename 
      EpsilonEquivalenceIteration<vector0<>, 
                                  typename push_back<vector0<>, _State>::type, 
                                  _Transitions>::type type;
    };
    
  }

}

#endif
