#ifndef FSM_BITS_ORDERED_HPP
#define FSM_BITS_ORDERED_HPP

#include <boost/mpl/vector.hpp>

#include <boost/mpl/push_front.hpp>
#include <boost/mpl/deref.hpp>
#include <boost/mpl/begin.hpp>
#include <boost/mpl/end.hpp>

#include <boost/mpl/if.hpp>
#include <boost/mpl/logical.hpp>
#include <boost/mpl/comparison.hpp>

// Rendezett beszuras es unio muvelet.

namespace fsm {

  namespace bits {

  using namespace boost;
  using namespace boost::mpl;

    template <typename _Begin, typename _End>
    struct ordered_insert_iter_no_value {
      typedef typename
      push_front<typename ordered_insert_iter_no_value<
                   typename next<_Begin>::type, _End>::type, 
                 typename deref<_Begin>::type>::type type;
    };

    template <typename _End>
    struct ordered_insert_iter_no_value<_End, _End> {
      typedef vector0<> type;
    };

    template <typename _Begin, typename _End, 
              typename _Value, typename _Less>
    struct ordered_insert_iter {
      typedef typename 
      if_<typename is_same<typename deref<_Begin>::type, _Value>::type,
          typename ordered_insert_iter_no_value<_Begin, _End>::type,
          typename if_<
            typename _Less::template apply<typename deref<_Begin>::type, 
                                           _Value>::type,
            typename push_front<
              typename ordered_insert_iter_no_value<_Begin, _End>::type, 
              _Value>::type,
            typename push_front<
              typename ordered_insert_iter<
                typename next<_Begin>::type, _End, _Value, _Less>::type,
              typename deref<_Begin>::type>::type 
            >::type
      >::type type;

    };
    
    template <typename _End, typename _Value, typename _Less>
    struct ordered_insert_iter<_End, _End, _Value, _Less> {
      typedef typename 
      push_front<vector0<>, _Value>::type type;
      
    };

    template <typename _Vector, typename _Value, typename _Less>
    struct ordered_insert {

      typedef typename ordered_insert_iter<
        typename begin<_Vector>::type, 
        typename end<_Vector>::type, _Value, _Less>::type type;
    };

    template <typename _LeftBegin, typename _LeftEnd,
              typename _RightBegin, typename _RightEnd,
              typename _Less> 
    struct ordered_merge_iter {
      
      typedef typename 
      if_<typename is_same<typename deref<_LeftBegin>::type, 
                           typename deref<_RightBegin>::type>::type,
          typename push_front<
            typename ordered_merge_iter<
              typename next<_LeftBegin>::type, _LeftEnd, 
              typename next<_RightBegin>::type, _RightEnd, _Less>::type, 
            typename deref<_LeftBegin>::type>::type,
          typename if_<
            typename _Less::template apply<
              typename deref<_LeftBegin>::type, 
              typename deref<_RightBegin>::type>::type, 
                typename push_front<
                typename ordered_merge_iter<
                  _LeftBegin, _LeftEnd, 
                  typename next<_RightBegin>::type, _RightEnd, 
                  _Less>::type, 
                    typename deref<_RightBegin>::type>::type,
                      typename push_front<
                      typename ordered_merge_iter<
                        typename next<_LeftBegin>::type, _LeftEnd, 
                        _RightBegin, _RightEnd, _Less>::type,
                        typename deref<_LeftBegin>::type>::type
            >::type 
      >::type type;

    };

    template <typename _LeftEnd,
              typename _RightBegin, typename _RightEnd,
              typename _Less> 
    struct ordered_merge_iter<_LeftEnd, _LeftEnd, 
                              _RightBegin, _RightEnd, _Less> {
      typedef typename 
      push_front<typename ordered_merge_iter<
                   _LeftEnd, _LeftEnd, 
                   typename next<_RightBegin>::type, _RightEnd, _Less>::type, 
                 typename deref<_RightBegin>::type>::type type;
    };
      
    template <typename _LeftBegin, typename _LeftEnd,
              typename _RightEnd,
              typename _Less> 
    struct ordered_merge_iter<_LeftBegin, _LeftEnd, 
                              _RightEnd, _RightEnd, _Less> {
      typedef typename 
      push_front<typename ordered_merge_iter<
                   typename next<_LeftBegin>::type, _LeftEnd,
                   _RightEnd, _RightEnd, _Less>::type, 
                 typename deref<_LeftBegin>::type>::type type;
    };

    template <typename _LeftEnd,
              typename _RightEnd,
              typename _Less> 
    struct ordered_merge_iter<_LeftEnd, _LeftEnd, 
                              _RightEnd, _RightEnd, _Less> {
      typedef vector0<> type;
    };

    template <typename _LeftVector, typename _RightVector, typename _Less>
    struct ordered_merge {
      typedef typename ordered_merge_iter<
        typename begin<_LeftVector>::type,
        typename end<_LeftVector>::type,
        typename begin<_RightVector>::type,
        typename end<_RightVector>::type,
        _Less>::type type;
    };
    
  }
}

#endif
