#ifndef FSM_FSM_HPP
#define FSM_FSM_HPP

#include <boost/mpl/vector.hpp>
#include <boost/mpl/push_back.hpp>
#include <boost/mpl/copy.hpp>

#include <fsm/state.hpp>
#include <fsm/transition.hpp>

#include <fsm/bits/ordered.hpp>
#include <fsm/bits/indicators.hpp>
#include <fsm/bits/fsm_impl.hpp>
#include <fsm/bits/available_states.hpp>

#include <boost/mpl/placeholders.hpp>

// A veges allapotu automata. A _Start egy tetszoleges allapot tipus,
// a _StopSet egy tetszoleges mpl sorozat allapot tipusokbol. A
// _Transitions az atmeneteket irja le, egy mpl sorozat olyan
// tipusokbol, melyeknek van egy Source allapot tipusa, egy Target
// allapot tipusa es egy mpl integral_c karakter tipus. A _StateLess
// egy teljes rendezes metafuggveny osztaly az allapot
// tipusokon. Allapot tipusnak erdemes a State tipust hasznalni, mig
// az atmeneteknek a Transition tipust.
namespace fsm {

  template <typename _Start, typename _StopSet, typename _Transitions, 
            typename _StateLess = StateLess_ >
  struct FiniteStateMachine {

    static bool check(const char *ptr) {
      typedef typename bits::EpsilonEquivalence<_Start, 
        _Transitions>::type StartSet;
      typedef typename boost::mpl::fold<_StopSet, boost::mpl::vector0<>, 
        bits::ordered_insert<boost::mpl::placeholders::_1, 
        boost::mpl::placeholders::_2, _StateLess> >::type StopSet;
      typedef typename bits::AvailableTransitions<StopSet, 
        _Transitions>::type Transitions;
      return bits::Step<StartSet, StopSet, 
        Transitions, _StateLess>::check(ptr);
    }

  };
  
}

#endif
